/*
 *  BINtegrity aka Syswall
 *
 */

#ifdef __KERNEL__
#include <linux/types.h>
#else
#include <stdio.h>
#include <stdlib.h>
#include <stdint.h>
#endif

#include "mips_dasm.h"

#define BITMASK(n) ((1 << n) - 1)

// for ops look at:
// https://www.student.cs.uwaterloo.ca/~isg/res/mips/opcodes

#define CHECK(SRC) if (r[instruction->src[SRC]] == 0xdeadbeef) {r[instruction->dst] = 0xdeadbeef; return -2;}

int mips_execute(mips_instruction_t* instruction, reg_state_t* reg_state) {
    int32_t *r = reg_state->regs;
    switch (instruction->op) {
        case NOP:
            return 0;

        case ADDU:
            CHECK(0);
            CHECK(1);
            r[instruction->dst] = r[instruction->src[0]] + r[instruction->src[1]];
            return 0;
            
        case ADDIU:
            CHECK(0);
            r[instruction->dst] = r[instruction->src[0]] + instruction->imm;    
            return 0;
            
        case LUI:
            r[instruction->dst] = instruction->imm << 16;
            return 0;
            
        case SLL:
            CHECK(0);
            r[instruction->dst] = r[instruction->src[0]] << instruction->imm;
            return 0;

        case SRA:
            CHECK(0);
            r[instruction->dst] = r[instruction->src[0]] >> instruction->imm;
            return 0;

        default:
            r[instruction->dst] = 0xdeadbeef;
            return -1;
    }
    return -3;
}

int mips_disassemble(mips_instruction_t* instruction, uint32_t number) {
    // In the MIPS architecture, all machine instructions are represented as 32-bit numbers
    uint8_t op, op_lower, op_upper,
            rs, rd, 
            rt, rt_upper, rt_lower, 
            sa, 
            funct, funct_upper, funct_lower;
    int16_t imm;
    int32_t target;

    op = number >> 26;
    op_upper = (op >> 3) & BITMASK(3);
    op_lower = op & BITMASK(3);

    rs = (number >> 21) & BITMASK(5);

    rt = (number >> 16) & BITMASK(5);
    rt_upper = rt >> 3;
    rt_lower = rt & BITMASK(3);

    rd = (number >> 11) & BITMASK(5);
    sa = (number >> 6) & BITMASK(5);

    funct = number & BITMASK(6);
    funct_upper = (funct >> 3) & BITMASK(3);
    funct_lower = funct & BITMASK(3);

    imm = number & BITMASK(16);

    // Tricky business: two's complement 26-bit
    target = number & BITMASK(26);
    if(target > BITMASK(25)) {
        target |= 0xfc000000;
    }

    instruction->type = UNKNOWN;
    
    if(number == 0) {
        instruction->op = NOP;
    } else if(op == 0) {
        instruction->op = mips_register_instruction_map[funct_upper][funct_lower];
        switch(funct_upper) {
        case MIPS_REG_TYPE_SHIFT_OR_SHIFTV:
            instruction->type = ARTH_LOG;
            if(funct_lower < 4) {
                // sll	000000	Shift	$d = $t << sa
                // sra	000011	Shift	$d = $t >> a
                // srl	000010	Shift	$d = $t >>> a
                instruction->dst = rd;
                instruction->src[0] = rt;
                instruction->imm = sa;
                //sprintf(instruction->arguments, "%s, %s, %d", MIPS_REGISTER_NAMES[rd], MIPS_REGISTER_NAMES[rt], sa);
            } else {
                // sllv	000100	ShiftV	$d = $t << $s
                // srav	000111	ShiftV	$d = $t >> $s
                // srlv	000110	ShiftV	$d = $t >>> $s
                instruction->dst = rd;
                instruction->src[0] = rt;
                instruction->src[1] = rs;
                //sprintf(instruction->arguments, "%s, %s, %s", MIPS_REGISTER_NAMES[rd], MIPS_REGISTER_NAMES[rt], MIPS_REGISTER_NAMES[rs]);
            }
            break;

        case MIPS_REG_TYPE_JUMPR:
            instruction->type = JUMP;
            if(funct_lower < 1) {
                // jr	001000	JumpR	pc = $s
                instruction->src[0] = rs;
                //sprintf(instruction->arguments, "%s", MIPS_REGISTER_NAMES[rs]);
            } else {
                // jalr	001001	JumpR	$ra = pc; pc = $s
                instruction->dst = rd;
                instruction->src[0] = rs;
                //sprintf(instruction->arguments, "%s, %s", MIPS_REGISTER_NAMES[rd], MIPS_REGISTER_NAMES[rs]);
            }
            break;

        case MIPS_REG_TYPE_MOVE:
            if(funct_lower % 2 == 0) {
                // mfhi	010000	MoveFrom	$d = hi
                // mflo	010010	MoveFrom	$d = lo
                instruction->dst = rd;
                //sprintf(instruction->arguments, "%s", MIPS_REGISTER_NAMES[rd]);
            } else {
                // mthi	010001	MoveTo	hi = $s
                // mtlo	010011	MoveTo	lo = $s
                instruction->src[0] = rs;
                //sprintf(instruction->arguments, "%s", MIPS_REGISTER_NAMES[rs]);
            }
            break;

        case MIPS_REG_TYPE_DIVMULT:
            instruction->type = ARTH_LOG;
            // div	011010	DivMult	lo = $s / $t; hi = $s % $t
            // divu	011011	DivMult	lo = $s / $t; hi = $s % $t
            // mult	011000	DivMult	hi:lo = $s * $t
            // multu	011001	DivMult	hi:lo = $s * $t
            instruction->src[0] = rs;
            instruction->src[1] = rt;
            //sprintf(instruction->arguments, "%s, %s", MIPS_REGISTER_NAMES[rs], MIPS_REGISTER_NAMES[rt]);
            break;

        case MIPS_REG_TYPE_ARITHLOG_GTE:
        case MIPS_REG_TYPE_ARITHLOG_GTE + 1:
            instruction->type = ARTH_LOG;
            // sub	100010	ArithLog	$d = $s - $t
            // subu	100011	ArithLog	$d = $s - $t
            // add	100000	ArithLog	$d = $s + $t
            // addu	100001	ArithLog	$d = $s + $t
            // xor	100110	ArithLog	$d = $s ^ $t
            // and	100100	ArithLog	$d = $s & $t
            instruction->dst = rd;
            instruction->src[0] = rs;
            instruction->src[1] = rt;
            //sprintf(instruction->arguments, "%s, %s, %s", MIPS_REGISTER_NAMES[rd], MIPS_REGISTER_NAMES[rs], MIPS_REGISTER_NAMES[rt]);
            break;

        default:
            return 0;
        }

    } else if(op == 0x1c) {
        instruction->op = mips_register_c_instruction_map[funct_upper][funct_lower];
        switch(funct_upper) {
        case MIPS_REG_C_TYPE_MULTIPLY:
            instruction->type = ARTH_LOG;
            if(funct_lower == 2) {
                instruction->dst = rd;
                instruction->src[0] = rs;
                instruction->src[1] = rt;
                //sprintf(instruction->arguments, "%s, %s, %s", MIPS_REGISTER_NAMES[rd], MIPS_REGISTER_NAMES[rs], MIPS_REGISTER_NAMES[rt]);
            } else {
                instruction->src[0] = rs;
                instruction->src[1] = rt;
                //sprintf(instruction->arguments, "%s, %s", MIPS_REGISTER_NAMES[rs], MIPS_REGISTER_NAMES[rt]);
            }
            break;

        case MIPS_REG_C_TYPE_COUNT:
            instruction->dst = rd;
            instruction->src[0] = rs;
            //sprintf(instruction->arguments, "%s, %s", MIPS_REGISTER_NAMES[rd], MIPS_REGISTER_NAMES[rs]);
            break;
        }

    } else if(op == 1) {
        instruction->op = mips_register_rt_instruction_map[rt_upper][rt_lower];
        instruction->dst = rs;
        instruction->imm = imm;
        if (rt_upper == 0 || rt_upper == 2) {
            instruction->type = JUMP;
        }
        //sprintf(instruction->arguments, "%s, %d", MIPS_REGISTER_NAMES[rs], imm);
    } else {
        instruction->op = mips_root_instruction_map[op_upper][op_lower];
        switch(op_upper) {
        case MIPS_ROOT_TYPE_JUMP_OR_BRANCH:
            instruction->type = JUMP;
            if(op_lower < 4) {
                // j	000010	Jump	pc += i << 2
                // jal	000011	Jump	$ra = pc; pc += i << 2
                instruction->dst = RA;
                instruction->imm = target;
                //sprintf(instruction->arguments, "%d", target);
            } else {
                if(op_lower < 6) { //Branch
                    // beq	000100	Branch	if ($s == $t) pc += imm << 2
                    // bne	000101	Branch	if ($s != $t) pc += i << 2
                    instruction->src[0] = rs;
                    instruction->src[1] = rt;
                    instruction->imm = imm;
                    //sprintf(instruction->arguments, "%s, %s, %d", MIPS_REGISTER_NAMES[rs], MIPS_REGISTER_NAMES[rt], imm);
                } else {
                    // bgtz	000111	BranchZ	if ($s > 0) pc += imm << 2
                    // blez	000110	BranchZ	if ($s <= 0) pc += imm << 2
                    instruction->src[0] = rs;
                    instruction->imm = imm;
                    //sprintf(instruction->arguments, "%s, %d", MIPS_REGISTER_NAMES[rs], imm);
                }
            }
            break;

        case MIPS_ROOT_TYPE_ARITHLOGI:
            instruction->type = ARTH_LOG;
            if(op_lower < 7) { // ??? < 7 == < 111
                // addi	001000	ArithLogI	$t = $s + SE(imm)
                // addiu	001001	ArithLogI	$t = $s + SE(imm)
                // slti	001010	ArithLogI	$t = ($s < SE(imm))
                // sltiu	001001	ArithLogI	$t = ($s < SE(imm))
                // andi	001100	ArithLogI	$t = $s & ZE(imm)
                // ori	001101	ArithLogI	$t = $s | ZE(imm)
                // xori	001110	ArithLogI	$d = $s ^ ZE(imm)
                instruction->dst = rt;
                instruction->src[0] = rs;
                instruction->imm = imm;
                //printf("%s, %s, %d", mips_register_names[rt], mips_register_names[rs], imm);
            } else {
                instruction->dst = rt;
                instruction->imm = imm;
                //printf("%s, %d\n", mips_register_names[rt], imm);
            }
            break;

        // TODO: distinguish between load and store for dst/src distinction!
        case MIPS_ROOT_TYPE_LOADSTORE_GTE:
        case MIPS_ROOT_TYPE_LOADSTORE_GTE+1:
        case MIPS_ROOT_TYPE_LOADSTORE_GTE+2:
        case MIPS_ROOT_TYPE_LOADSTORE_GTE+3:
            // lb	100000	LoadStore	$t = SE (MEM [$s + i]:1)
            // lbu	100100	LoadStore	$t = ZE (MEM [$s + i]:1)
            // lh	100001	LoadStore	$t = SE (MEM [$s + i]:2)
            // lhu	100101	LoadStore	$t = ZE (MEM [$s + i]:2)
            // lw	100011	LoadStore	$t = MEM [$s + i]:4
            // sb	101000	LoadStore	MEM [$s + i]:1 = LB ($t)
            // sh	101001	LoadStore	MEM [$s + i]:2 = LH ($t)
            // sw	101011	LoadStore	MEM [$s + i]:4 = $t
            instruction->dst = rt;
            instruction->src[0] = rs;
            instruction->imm = imm;
            //sprintf(instruction->arguments, "%s, %d(%s)", MIPS_REGISTER_NAMES[rt], imm, MIPS_REGISTER_NAMES[rs]);
            break;
        

        default:
            return 0;
        }
    }

    return 1;
}
